/*
 * -------------------------------------------------------
 *
 *  Copyright (C) 1997 1999 2000 Jochen Karrer
 *
 *  Released under the terms of the GPL.
 *
 * -------------------------------------------------------
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/io.h>

#define MAXBOARDS 1
static unsigned int adc_iobase[MAXBOARDS]={0x360,};
static unsigned int adc_irq[MAXBOARDS]={0x00,};
MODULE_PARM(adc_iobase,"1-" __MODULE_STRING(MAXBOARDS) "i");
MODULE_PARM(adc_irq,"1-" __MODULE_STRING(MAXBOARDS) "i");


#define COMMAND_PORT(adc) ((adc)->iobase+4)
#define INPUT_PORT(adc) ((adc)->iobase+5)
#define DATA_PORT(adc) ((adc)->iobase+6)
#define C54_CONTROL_PORT(adc) ((adc)->iobase+3)
#define C54_COUNTER(counter,adc) ((adc)->iobase+(counter))

// 60 is MAJOR for experimental use 
#ifndef ADC_MAJOR
  #define ADC_MAJOR 60 
#endif
#define N_CCD_PIXELS (2048)
#define FIFOSIZE (12*N_CCD_PIXELS)
#define MAX_AQUISITIONS_PER_SEC 20
#define MAX_CMDLEN 40

#define EOF_SENDABLE   0
#define EOF_PENDING    1
#define EOF_SENT       2

struct Adc500_s;
static int mod_use_count=0;
static void s2000_set_integration_time(int time,struct Adc500_s *adc);
static void s2000_trigger_aquisition(struct Adc500_s *adc,int channel);
static int s2000_init(struct Adc500_s *adc);

typedef struct {
	int (*init)(struct Adc500_s *adc);
	void (*set_integration_time)(int time,struct Adc500_s *adc);
	void (*trigger_aquisition)(struct Adc500_s *adc,int channel);
	int  n_pixels;
} spectro_ops;

static spectro_ops 
s2000_spec_ops = {
	s2000_init,
	s2000_set_integration_time,
	s2000_trigger_aquisition,
	2048,
};

typedef struct  Adc500_s {
	unsigned int iobase;	
	unsigned int irq;

	// statistics
	unsigned int n_irqs;
	unsigned int watchdog_time; // for watchdog

	// Software fifo
	unsigned int * fifo;
	unsigned int fifo_wp;	
	
	unsigned int message_wp;
	wait_queue_head_t wait_on_message;

	unsigned int continuous;
	unsigned int grab_done; // no further grabs until user gives command
	unsigned int add_frames;
	unsigned int framecounter;
	unsigned int channel;
	unsigned int integration_time;

	wait_queue_head_t wait_on_data;
	struct timer_list trigger_timer;
	int timer_busy;

	// register backingstores
	unsigned char cmd;

	spectro_ops *spec_ops;
} Adc500;

typedef struct {
	Adc500  *adc;
	unsigned int message_rp;
	unsigned int fifo_rp;
	char str[25];
	int (*readfunc)(struct file *,char *,int);
	unsigned int str_readpos;
	int  eof;

	// interpreter
	char cmd[MAX_CMDLEN];
	unsigned int  cmd_wp;

	// status 
	char *statuspage;
	int statuslen;
} adc_file_data;

static Adc500 *adcinfo[MAXBOARDS]={0,};

static void
sleep_schedule(unsigned long time)
{
	unsigned long timeout=jiffies+time;
	while (jiffies<timeout)
                schedule();
}

static inline void 
adc_load_counter(unsigned int counter,int value,Adc500 *adc) {
	int ctrl;
	if(counter > 2){
		printk("shit, trying to load nonexistent counter %d\n",counter);
		return;
	}
	ctrl = (counter << 6)+0x36;
	outb(ctrl,C54_CONTROL_PORT(adc));
	outb(value & 0xff,C54_COUNTER(counter,adc));
	outb((value>>8)&0xff,C54_COUNTER(counter,adc));
}

static inline void 
adc_reset_fifo(Adc500 *adc){
	adc->cmd |= 0x20;
	outb(adc->cmd,COMMAND_PORT(adc));
	adc->cmd &= ~0x20;
	outb(adc->cmd,COMMAND_PORT(adc));
}

static void 
s2000_set_integration_time(int time,Adc500 *adc) {
	if((time < 5) || (time>65535) ) 
		time = 5;
	adc->integration_time = time;
	adc_load_counter(2,time,adc);
}

static void 
s2000_trigger_aquisition(Adc500 *adc,int channel) {
	adc->watchdog_time = jiffies;
	adc->cmd &= ~0x98; 	// mask channel bits
	adc->cmd |= ((channel&3)<<3) + ((channel&4)<<5);
	adc->cmd &= ~0x41; 	// disable interrupt, disable trigger
	adc->cmd |= 0x20; 	// reset fifo
	outb(adc->cmd,COMMAND_PORT(adc));
	adc->cmd &= ~0x20; 	// stop reset 
	outb(adc->cmd,COMMAND_PORT(adc));
	adc->cmd |= 0x41; 	//  enable interrupt, enable trigger 
	adc->cmd |= 0x2; 	//  not dark, not external 
	outb(adc->cmd,COMMAND_PORT(adc));
	adc->cmd &= ~0x2;
}

static inline void 
adc_watchdog(Adc500 *adc) {
	if(adc->continuous && (adc->watchdog_time+2*adc->integration_time<jiffies)) {
		adc->watchdog_time = jiffies;
		adc->spec_ops->trigger_aquisition(adc,adc->channel);
	}
}

static inline void
align_read_fifo(Adc500 *adc) {
        u_int i = adc->fifo_wp  % N_CCD_PIXELS;
        adc->fifo_wp = ( adc->fifo_wp - i ) % FIFOSIZE;
}

static void
adc_timer_func(unsigned long data) {
	Adc500 *adc = (Adc500 *)data;
	adc->spec_ops->trigger_aquisition(adc,adc->channel);
	adc->timer_busy = 0;
}

/*
 * -------------------------------------------------------------------
 *
 *  adc_interrupt
 * 
 *	read fifo and reset it, restart aquisition if continuous
 * 
 * -------------------------------------------------------------------
 */

static void 
adc_interrupt(int irq,void *dev,struct pt_regs *regs) {
	int i;
	Adc500 *adc = (Adc500 *)dev;
	static int intwatch; 
	static u_int last_jiffies;

	adc->n_irqs++;

	adc->cmd &= ~0x41; // turn interrupts and triggers off
	outb(adc->cmd,COMMAND_PORT(adc));
	sti();
	for(i=0;i<N_CCD_PIXELS;i++){
		unsigned int value = inw(DATA_PORT(adc));
		if(adc->framecounter == adc->add_frames) 
			adc->fifo[(adc->fifo_wp+i)%FIFOSIZE] = value;
		else 
			adc->fifo[(adc->fifo_wp+i)%FIFOSIZE] += value;
	}	
	adc_reset_fifo(adc);
	if(jiffies > last_jiffies)
		intwatch = 0;
	intwatch++;
	if(adc->continuous) {
		if(intwatch > 300 / HZ) {       // to much interrupts
			adc->continuous = 0;    // disable one board
		}
		else {
			cli();
			if(adc->integration_time>1000/MAX_AQUISITIONS_PER_SEC){ 
				sti();
				adc->spec_ops->trigger_aquisition(adc,adc->channel);	
			} else if(!adc->timer_busy) {
				init_timer(&adc->trigger_timer);
				adc->trigger_timer.expires = jiffies+HZ/MAX_AQUISITIONS_PER_SEC;
				add_timer(&adc->trigger_timer);
				adc->timer_busy = 1;
			}
			sti();
		}
	} else if(intwatch > 300/HZ) {	// still interrupts ?
		free_irq(irq,dev);  	// ( wrong ioport, correct irq for example)
		adc->irq = 0;
	}
	adc->framecounter--;
	if(adc->framecounter == 0){
		if (adc->continuous == 0) 
			adc->grab_done = 1;  
		adc->framecounter = adc->add_frames;
		adc->fifo_wp = (adc->fifo_wp+N_CCD_PIXELS)%FIFOSIZE;
		align_read_fifo(adc);
		wake_up_interruptible(&adc->wait_on_data);
	}
	return;
}

/*
 * --------------------------------------------------------------------------
 *
 * 	adc_llseek -
 *
 *		SEEK_END: skip data in read-buffer
 *
 * -------------------------------------------------------------------------
 */

static loff_t 
adc_llseek(struct file * file, loff_t offset, int orig){
	adc_file_data *pd = (adc_file_data *) file->private_data;

	if (orig == 2) { // SEEK_END
		pd->fifo_rp = pd->adc->fifo_wp;
                return file->f_pos;
	}
	return file->f_pos;
}

/*
 * ---------------------------------------------------------------------------
 *
 * 	adc_binary_read -
 *
 *		puts the data to user space in binary (unsigned int) format 
 *
 * ---------------------------------------------------------------------------
 */
static int 
adc_binary_read(struct file *file,char *buf,int count)
{
	adc_file_data *pd = (adc_file_data *) file->private_data;
	Adc500 *adc = pd->adc;
	int i;
	if(pd->eof == EOF_PENDING ) {
		pd->eof = EOF_SENT;
		return 0;
	}
	count = count - (count % sizeof(*adc->fifo));  // align
	for(i=0;i<count;i+=sizeof(*adc->fifo)) {
		if((pd->eof !=EOF_SENT) && ((pd->fifo_rp % N_CCD_PIXELS)==0))  {
			if(i) 
				pd->eof = EOF_PENDING;
			else 	
				pd->eof = EOF_SENT;
			return i;
		}
		while(adc->fifo_wp == pd->fifo_rp) {
			if(i)
				return i;
			if(file->f_flags & O_NONBLOCK) 
				return -EWOULDBLOCK;
			interruptible_sleep_on(&adc->wait_on_data);
			if(signal_pending(current)) {
				return -ERESTARTSYS;
			}
		}	
		put_user(adc->fifo[pd->fifo_rp],(typeof(adc->fifo))(buf+i));
		pd->fifo_rp = (pd->fifo_rp+1) % FIFOSIZE;	
		pd->eof = EOF_SENDABLE;
		file->f_pos += sizeof(*adc->fifo);
	}
	return i;	
}

/*
 * ----------------------------------------------------------------------
 *
 * 	adc_ascii_read
 *
 *		put the data in ascii format to user space
 *
 * ----------------------------------------------------------------------
 */
static int
adc_ascii_read(struct file *file,char *buf,int count) {
	adc_file_data *pd = (adc_file_data *) file->private_data;
	Adc500 *adc = pd->adc;
	int i;
	if(pd->eof == EOF_PENDING ) {
		pd->eof = EOF_SENT;
		return 0;
	}
	for(i=0;i<count;i++) {
		if(pd->str[pd->str_readpos] == 0) {
			if((pd->eof !=EOF_SENT) && ((pd->fifo_rp % N_CCD_PIXELS)==0))  {
				if(i) 
					pd->eof = EOF_PENDING;
				else 	
					pd->eof = EOF_SENT;
				return i;
			}
			while(adc->fifo_wp == pd->fifo_rp) {
				if(i)
					return i;
				if(file->f_flags & O_NONBLOCK) 
					return -EWOULDBLOCK;
				interruptible_sleep_on(&adc->wait_on_data);
				if(signal_pending(current)) 
					return -ERESTARTSYS;
			}
			pd->str_readpos=0;
			sprintf(pd->str,"%u\n",adc->fifo[pd->fifo_rp]);
			pd->fifo_rp = (pd->fifo_rp+1) % FIFOSIZE;	
			pd->eof = EOF_SENDABLE;
			if(current->need_resched)
				schedule();
		}
		put_user(pd->str[pd->str_readpos],buf+i);
		pd->str_readpos++;
		file->f_pos++;			
	}
	return i;
}


static ssize_t 
adc_read(struct file *file,char *buf,size_t count,loff_t *ppos) 
{
	adc_file_data *pd = (adc_file_data *) file->private_data;
	int error = verify_area(VERIFY_WRITE,buf,count);
        if (error)
                return error;
	return pd->readfunc(file,buf,count);
}

/*
 * --------------------------------------
 * 
 * sscan an integer
 *
 * --------------------------------------
 */
static void 
adc_isscan(char *str,u_int *value) {
	int i=0;
	*value = 0;
	while(str[i]==' ')
		i++;
	while((str[i]>='0') && (str[i]<='9')) {
		*value = *value * 10 + str[i] - '0';
		i++;
	}	
}
/*
 * -----------------------------------------------------------------
 *
 *  adc_interpreter -
 *
 *	interprete the ascii strings sent to /dev/adc500 
 *
 * -----------------------------------------------------------------
 */
static void
adc_interpreter(char *cmd,adc_file_data *pd) 
{
	u_int time;
	Adc500 *adc = pd->adc;
	adc_isscan(cmd,&time);
	if(time) {
		adc->spec_ops->set_integration_time(time,adc);
		return;
	}
	switch (cmd[0]) {
		case 'a':			
			if(cmd[1] == 'd') {	 // add 
				adc_isscan(cmd+3,&adc->add_frames);
				if(adc->add_frames <=0) 
					adc->add_frames = 1;
				adc->framecounter=1;
			} else if (cmd[1]=='s') {   // ascii
				pd->readfunc = adc_ascii_read;
			}
			break;

		case 'b':
			pd->readfunc = adc_binary_read;
			break;
			
		case 'c':			// channel
			adc_isscan(cmd+7,&adc->channel);
			adc->channel &=7;
			break;

		case 's':
			if(cmd[1] == 'i') { 		// single spectrum
				adc->grab_done = 0;
				adc->continuous = 0;
				adc->framecounter = adc->add_frames;
				adc->spec_ops->trigger_aquisition(adc,adc->channel); 
			} else if(cmd[2] == 'a') {	 // start continuous
				adc->grab_done = 0;
				adc->continuous = 1;
				adc->framecounter = adc->add_frames;
				adc->spec_ops->trigger_aquisition(adc,adc->channel); 
			} else if (cmd[2] == 'o')  	// stop
				adc->continuous = 0;
			break;
	} 
}

static ssize_t 
adc_write(struct file *file,const char *buf,size_t count,loff_t *ppos) 
{
	adc_file_data *pd = (adc_file_data *) file->private_data;
	int i;
	int error = verify_area(VERIFY_READ,buf,count);
	if(error)
		return error;

	for(i=0;i<count;i++) {
		char c;
		get_user(c,&buf[i]);
		if ( pd->cmd_wp < MAX_CMDLEN-1  ) {
			pd->cmd[pd->cmd_wp]=c;
			pd->cmd[pd->cmd_wp+1]=0;
			pd->cmd_wp++;
		} 
		if( c == '\n') {
			adc_interpreter(pd->cmd,pd);
			pd->cmd_wp = 0;
		}
	}
	*ppos+=i;
	return i;
}

static unsigned int 
adc_poll(struct file *file, poll_table *wait) {
	adc_file_data *pd = (adc_file_data *) file->private_data;
	Adc500 *adc = pd->adc;
	unsigned int mask = 0;
	poll_wait(file,&adc->wait_on_data,wait);
	if(pd->fifo_rp != adc->fifo_wp) {
		mask |= POLLIN | POLLRDNORM;
	}
	return mask;
}

static int 
adc_release(struct inode *inode,struct file *file) {
	adc_file_data *pd = (adc_file_data *) file->private_data;
	kfree(pd);
	MOD_DEC_USE_COUNT;
	mod_use_count--;
	return 0;
}

static int
adc_open( struct inode* inode, struct file* filp)
{
        adc_file_data *pd;
	int minor =  MINOR(inode->i_rdev);
	if((minor >= MAXBOARDS) || !(adcinfo[minor]) ) {
		return -ENODEV;
	}
        if(!(filp->private_data =
            kmalloc(sizeof(adc_file_data),GFP_KERNEL)))
                return -ENOMEM;
	pd = (adc_file_data *) filp->private_data;
	pd->cmd_wp=0;
	pd->adc = adcinfo[minor];
	if (pd->adc->grab_done) {
		pd->fifo_rp = (pd->adc->fifo_wp - N_CCD_PIXELS) % FIFOSIZE;
	} else {
		pd->fifo_rp = pd->adc->fifo_wp;
	}
	pd->readfunc = adc_ascii_read;
	pd->str[0]=0;
	pd->str_readpos=0;
	pd->eof = EOF_SENT;
	adc_watchdog(pd->adc);
	MOD_INC_USE_COUNT;
	mod_use_count++;
	return 0;
}

static struct file_operations adc_fops = {
	llseek:	adc_llseek,  
	read:	adc_read,  
	write:  adc_write, 
	poll:   adc_poll,  
	open:   adc_open, 
	release:adc_release,
};

/*
 * ----------------------------------------------------------------
 *
 * Proc dir entry for the s2000 
 *
 *-----------------------------------------------------------------
 */


static int 
sprint_status(char *page) {
	char *str = page;
	Adc500 *adc;
	int i;
	for(i=0;i<MAXBOARDS;i++) {
		adc=adcinfo[i];
		str+=sprintf(str,"\nS2000/ADC-500 Board %d Status:\n",i);
		str+=sprintf(str,"---------------------------------\n");
		str+=sprintf(str,"State: ");	
		if(!adc->continuous  && adc->grab_done) 
			str+=sprintf(str,"Aquisition is stopped\n");	
		else if (adc->continuous == 0 && !adc->grab_done) {
			str+=sprintf(str,"Waiting for single-shot\n");	
		} else {
			str+=sprintf(str,"Continuous aquisition\n");	
		}
		str+=sprintf(str,"IO-Baseaddress:      0x%x\n",adc->iobase);
		str+=sprintf(str,"Interrupt:           %d\n",adc->irq);
		str+=sprintf(str,"received Interrupts: %d\n",adc->n_irqs);
		str+=sprintf(str,"active channel:      %d\n",adc->channel);
		str+=sprintf(str,"Integration Time:    %d ms\n",adc->integration_time);	
		str+=sprintf(str,"Frames to add:       %d\n" , adc->add_frames);
		str+=sprintf(str,"Number Of Users:     %d\n" , mod_use_count);
	}
	str[0]=0;	
	return str-page;
}

/*
 * ------------------------------------------------------------------------
 *
 *  adc_proc_read --
 *
 * 	print the state of one board into a string, and put
 *	the string to userspace
 *
 * ------------------------------------------------------------------------
 */
static ssize_t 
adc_proc_read(struct file * file,char *buf,size_t count,loff_t *fpos){
       adc_file_data *pd = (adc_file_data *) file->private_data;
	if(file->f_pos == 0) {
		char *str = pd->statuspage;
		str += sprint_status(str);
		pd->statuslen = str-pd->statuspage;
	}
	if(file->f_pos>=pd->statuslen) {
		return 0;
	}
        if((file->f_pos+count)>pd->statuslen) {
                count = pd->statuslen - file->f_pos;
        }
        copy_to_user(buf,pd->statuspage+file->f_pos,count);
        file->f_pos+=count;
        return count;
}


static unsigned int 
adc_proc_poll(struct file *file, poll_table *wait) {
	adc_file_data *pd = (adc_file_data *) file->private_data;
	Adc500 *adc = pd->adc;
	unsigned int mask = 0;
	poll_wait(file,&adc->wait_on_message,wait);
	if(pd->message_rp != adc->message_wp) {
		mask |= POLLIN | POLLRDNORM;
	}
	return mask;
}

static int 
adc_proc_release(struct inode *inode,struct file *file) {
	adc_file_data *pd = (adc_file_data *) file->private_data;
	if(pd->statuspage)
		vfree(pd->statuspage);
	kfree(pd);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int
adc_proc_open( struct inode* inode, struct file* filp) {
	char *str;
	adc_file_data *pd;
	struct proc_dir_entry * dp;

        dp = (struct proc_dir_entry *) inode->u.generic_ip;
        if(!(filp->private_data =
            kmalloc(sizeof(adc_file_data),GFP_KERNEL)))
                return -ENOMEM;
        pd = (adc_file_data *) filp->private_data;
        pd->statuspage = vmalloc(PAGE_SIZE);
        if(pd->statuspage == NULL) {
		kfree(pd);
		return -ENOMEM;
        }
        pd->adc = adcinfo[0];
        pd->message_rp = pd->adc->message_wp;
        str = pd->statuspage;
        str += sprint_status(str);
        pd->statuslen = str-pd->statuspage;
	MOD_INC_USE_COUNT;
	return 0;
}

static loff_t
adc_proc_llseek(struct file * file, loff_t offset, int orig)
{
	adc_file_data *pd = (adc_file_data *) file->private_data;
    	switch (orig) {
	    case 0:
		file->f_pos = offset;
		return(file->f_pos);
	    case 1:
		file->f_pos += offset;
		return(file->f_pos);
	    case 2:
		file->f_pos=pd->statuslen;
		return(file->f_pos);
	    default:
		return(-EINVAL);
	}
}


static struct file_operations adc_proc_fops = {
	poll:		adc_proc_poll,
        llseek:         adc_proc_llseek,
        read:           adc_proc_read,
	open:		adc_proc_open,
	release:	adc_proc_release,
};

static 
void register_procinfo() {
	int i;
	char name[10];
	struct proc_dir_entry *dir;
	struct proc_dir_entry *ent;
	dir = create_proc_entry("adc500",S_IFDIR|S_IRUGO|S_IXUGO,0);
	for(i=0;i<MAXBOARDS;i++) {
		if(!adcinfo[i])
			continue;
		sprintf(name,"s2000_%d",i);
		ent = create_proc_entry(name,S_IFREG|S_IRUGO,dir);
		ent->proc_fops = &adc_proc_fops;
		ent->data = (void *)adcinfo[i];
	}
}

void unregister_procinfo(void) {
	int i;
	static char name[20];
	for(i=0;i<MAXBOARDS;i++) {
		sprintf(name,"adc500/s2000_%d",i);
		remove_proc_entry(name,0);
	}
	remove_proc_entry("adc500",0);
}

/*
 * -------------------------------------------------------------
 *
 *  cleanup_module()
 *
 *		free resources
 *
 * -------------------------------------------------------------
 */
static void 
cleanup_board(Adc500 *adc)
{
	unsigned long flags;

	adc->cmd &= ~0x41; // turn interrupts and triggers off
	outb(adc->cmd,COMMAND_PORT(adc));
	// this has to be done first !
	if(adc->irq)  {
		free_irq(adc->irq,adc);
		adc->irq=0;
	}
	save_flags(flags);
	cli();
	if(adc->timer_busy) { 
		del_timer(&adc->trigger_timer);
		adc->timer_busy=0;
	}
	restore_flags(flags);
	if(adc->iobase) {
		release_region(adc->iobase,0x8);
		adc->iobase=0;
	}
	if(adc->fifo != NULL ) { 
		vfree(adc->fifo);
		adc->fifo = NULL;
	}
}

void __exit
adc_cleanup_module(void) 
{
	int i;
	if(unregister_chrdev(ADC_MAJOR,"ADC-500") != 0) {
		printk("cleanup module failed\n");
		return;
	}
	unregister_procinfo();
	for(i=0;i<MAXBOARDS;i++)  {
		Adc500 *adc = adcinfo[i];
		if(adc) {
			cleanup_board(adc);
			kfree(adc);
			adcinfo[i] = NULL;
		}
	}	
}
/* 
 *  Counter 0: Master clock for the S2000 (2 * conversion frequency)
 *            programmed to 1 MHz
 *
 *  Counter 1: Connected to input of counter 2 which gives 
 *            the integration time and is set to 1ms 
 */

static int
s2000_init(Adc500 *adc) 
{
	adc_load_counter(0,8,adc);
	adc_load_counter(1,8000,adc);
	// probe if really a S2000
	// ...................
	return 1;
}

static int
probe_irq(Adc500 *adc) {
	unsigned long u;
	int irq;
	probe_irq_off(probe_irq_on());
	u = probe_irq_on();
	adc->spec_ops->set_integration_time(5,adc);
	adc->spec_ops->trigger_aquisition(adc,adc->channel); 
	sleep_schedule(HZ/10);
	irq = probe_irq_off(u);
	if ( irq > 0 ) {	
	printk("probed irq %d\n",irq);
		return irq;
	} 
	printk("autoprobing for irq failed\n");
	return 0;
}

/*
 * -----------------------------------------------------------------
 *
 *  init_module --
 *
 *	allocate resources and reset the card
 *
 * -------------------------------------------------------------------
 */

static Adc500 * 
init_board(Adc500 *adc,int i) 
{
	memset(adc,0,sizeof(Adc500));
	adc->spec_ops = &s2000_spec_ops;
	if(adc_iobase[i] == 0)
		return adc;
	init_waitqueue_head(&adc->wait_on_data); 
	init_waitqueue_head(&adc->wait_on_message); 
	if(request_region(adc_iobase[i],0x8,"ADC-500")==NULL) {
		printk("ADC-500: can't get IO-address 0x%x\n",adc_iobase[i]);
		return NULL;
	}
	adc->iobase = adc_iobase[i];

	// reset the hardware fifo before interrupts are enabled

	adc_reset_fifo(adc);

	// setup software fifo
	adc->fifo = (typeof(adc->fifo))vmalloc(FIFOSIZE*sizeof(*adc->fifo));
	if(adc->fifo==NULL) {
		return NULL;
	}

	// test if present with readback of control register
	inb(C54_COUNTER(0,adc));
	outb(0x80|0x40|0x20|0x02,C54_CONTROL_PORT(adc));
	if((inb(C54_COUNTER(0,adc))&0xf)!=0x6){
		printk("can't initialize counter! maybe ioport 0x%x is wrong\n",
			adc->iobase);
	}
	else 
		printk("found 8254, card seems to be present at 0x%x\n",
			adc->iobase);

	if(adc_irq[i] == 0) 
		adc_irq[i] = probe_irq(adc);
	if(adc_irq[i] == 0) {
		printk("probing for irq failed\n");
		return NULL;
	}

	if(request_irq(adc_irq[i], adc_interrupt,SA_INTERRUPT,"ADC-500",adc)){
		printk("can't get interrupt %d\n",adc_irq[i]);	
		return NULL;
	}
	adc->irq = adc_irq[i];

	if(adc->spec_ops->init(adc)) {
		printk("initialized ADC-500 (S2000) number %d\n",i);
	} else  {
		// try S1000
	}

	adc->trigger_timer.data = (unsigned long) adc;
	adc->trigger_timer.function = adc_timer_func;
	
	// try to trigger an interrupt:
	adc->spec_ops->set_integration_time(5,adc);
	adc->spec_ops->trigger_aquisition(adc,adc->channel); 
	sleep_schedule(HZ/10);
	if(adc->n_irqs<1) 
		printk("can't trigger interrupt %d\n",adc->irq);
	else 
		printk("triggering interrupt %d succeeded\n",adc->irq);
	adc->continuous = 1;
	adc->add_frames = 1;
	adc->framecounter = adc->add_frames;
	adc->spec_ops->set_integration_time(333,adc);
	adc->spec_ops->trigger_aquisition(adc,adc->channel); 
	return adc;
}

int __init
adc_init_module(void) 
{
	int i;
	if(register_chrdev(ADC_MAJOR,"ADC-500",&adc_fops)) {
		printk("register chardevice failed\n");
		return -EBUSY;
	}
	for(i=0;i<MAXBOARDS;i++) {
		Adc500 *adc = adcinfo[i] = kmalloc(sizeof(Adc500),GFP_KERNEL);
		if(adc==NULL) {
			adc_cleanup_module();
			return -ENOMEM;
		}
		memset(adc,0,sizeof(*adc));
		if(init_board(adc,i)==NULL) {
			printk("failed to init adc500: cleaning up %p\n",adc);
			cleanup_board(adc);
		}
	} 
	register_procinfo();
	return 0;
}

module_init(adc_init_module);
module_exit(adc_cleanup_module);
